"""
Relevance mapping module.

This module provides the interface for defining engine mappings.
"""

import copy
import typing


class Field(object):
    """
    Mapping field class.

    This class provides a generic interface for mapping a field name to a Python
    data type is engine agnostic.
    """
    def __init__(self, name: str, data_type: type, stored: bool=True, **options):
        """
        Initialize the field.

        :param name: the name of the field.
        :param data_type: the Python data type class.
        :param stored: whether to store the value of the field or just index it.
        :param options: additional options to store for the mapping.
        """
        self.name = name
        self.data_type = data_type
        self.stored = stored
        self.options = options

    def to_dict(self) -> dict:
        """
        Get the field as a dictionary.
        """
        data_type = '{0}.{1}'.format(
            self.data_type.__module__,
            self.data_type.__name__,
        )

        return {
            'name': self.name,
            'data_type': data_type,
            'stored': self.stored,
            'options': self.options,
        }

    def __repr__(self) -> str:
        """
        Overload operator for readable representation.
        """
        return '{0}:{1}'.format(self.name, self.data_type.__name__)


class ListField(Field):
    """
    List field class.

    This class provides the interface for mapping a list field. List fields are
    fields that contain multiple elements of a single defined type.
    """
    pass


class Mapping(object):
    """
    Mapping class.

    This class provides a generic interface for defining engine agnostic mappings.
    """
    def __init__(self, *fields):
        """
        Initialize the mapping.

        :param fields: the list of field objects for the mapping.
        """
        self.fields = list(fields)

    def add(self, field: Field) -> 'Mapping':
        """
        Add a field to the mapping.

        :param field: the field object to add.
        :returns: the new updated mapping object.
        """
        self = copy.deepcopy(self)
        self.fields.append(field)
        return self

    def to_dict(self) -> dict:
        """
        Get the mapping as a dictionary.
        """
        result = {}
        for x in self.fields:
            result[x.name] = x.to_dict()
        return result

    def __getitem__(self, name: str) -> Field:
        """
        Get a field by its name.

        :param name: the name of the field.
        :returns: the field object.
        """
        for x in self.fields:
            if x.name == name:
                return x
        raise KeyError(name)

    def __contains__(self, name: str) -> bool:
        """
        Check if a field is part of the mapping.

        :param name: the field name.
        :returns: True if it exists, False otherwise.
        """
        for x in self.fields:
            if x.name == name:
                return True
        return False

    def __iter__(self) -> typing.Iterable:
        """
        Iterate over all the fields.

        :returns: an iterator object.
        """
        return iter(self.fields)

    def __repr__(self) -> str:
        """
        Overload operator for readable representation.
        """
        return repr(list(self.fields))


class ObjectField(Field):
    """
    Object field class.

    This class provides the interface for mapping an object field. Object fields
    are fields that have nested mappings in them.
    """
    def __init__(self, name: str, mapping: Mapping, *args, **kwargs):
        """
        Initialize the field.

        :param name: the name of the field.
        :param mapping: the mapping object for the field.
        """
        super().__init__(name, dict, *args, **kwargs)
        self.mapping = mapping

    def to_dict(self) -> dict:
        """
        Get the field as a dictionary.
        """
        data = super().to_dict()
        del data['data_type']
        data['fields'] = self.mapping.to_dict()
        return data



"""
Relevance search REST API.

This module provides the REST API for search.
"""

import os
import pydoc
import anyconfig

from flask import request
from flask import jsonify
from flask import Response

from relevance.api import Flask
from relevance.api import start_app
from relevance.query import QueryParser
from relevance.query import QueryParserError
from relevance.facet import Facet
from relevance.engine import SearchEngine
from relevance.engine import MappingEngine


# Register the application
app = Flask('relevance.api.search')

# The configuration objects
app.data = type('app_data', (object,), dict(
    config={},
    engines={},
))()


def load_config(env_name: str=None):
    """
    Load or reload the configuration.

    :param env_name: the name of the environment to load. If omitted, it is taken
    from the PYTHON_ENV environment variable. If it doesn't exist, 'development' is
    assumed.
    """
    if env_name is None:
        env_name = os.environ.get('PYTHON_ENV', 'development')

    # Load the configuration file
    this_config = anyconfig.load(['./etc/default.json', './etc/{0}.json'.format(env_name)],
                                 ignore_missing=True)
    app.data.config.clear()
    app.data.config.update(this_config)

    # Load the engine instances
    this_engines = {}
    for engine_name, conf in app.data.config.items():
        try:
            cls_name = conf['engine']
            target = conf['target']
        except KeyError as ex:
            raise KeyError('missing configuration key {0}'.format(str(ex)))

        cls = pydoc.locate(cls_name)
        if cls is None or not issubclass(cls, SearchEngine):
            raise ImportError('cannot find engine {0}'.format(cls_name))

        facets = {}
        for facet_name, facet_conf in conf.get('facets', {}).items():
            facet_cls = pydoc.locate(facet_conf.get('type'))
            if facet_cls is None or not issubclass(facet_cls, Facet):
                raise ImportError('cannot find facet {0}'.format(facet_conf['type']))

            facet_obj = facet_cls(
                field=facet_conf.get('field'),
                path=facet_conf.get('path'),
                **facet_conf.get('options', {}),
            )

            facets[facet_name] = facet_obj

        options = conf.get('options', {})
        this_engines[engine_name] = cls(target, facets=facets, **options)

    app.data.engines.clear()
    app.data.engines.update(this_engines)


# Load the configuration initially
load_config()


@app.route('/', methods=['GET'])
def get_config() -> Response:
    """
    Get the configuration.
    """
    response = jsonify(app.data.config)
    response.status_code = 201
    return response


@app.route('/<engine_name>', methods=['GET'])
def do_search(engine_name: str) -> Response:
    """
    Perform a search request.

    Request arguments:
        q -- the query to run.

    .. seealso: :py:class:`QueryParser`

    :param engine_name: the name of the engine to use.

    Returns a result set object.
    """
    try:
        engine = app.data.engines[engine_name]
    except KeyError:
        response = jsonify({
            'error': {
                'desc': 'engine not found',
                'key': engine_name,
                'code': 'EngineNotFound',
            }
        })
        response.status_code = 404
        return response

    query = request.args.get('q', '')
    search = QueryParser().loads(query)

    results = engine.search(search)
    data = {
        'search': search.to_dict(),
        'results': results,
        'time': results.time,
        'count': results.count,
    }

    if results.facets is not None:
        data.update({
            'facets': results.facets,
        })

    response = jsonify(data)
    response.status_code = 200
    return response


@app.route('/<engine_name>/mapping', methods=['GET'])
def get_mapping_doc_types(engine_name: str) -> Response:
    """
    Get a list of available doc types.

    :param engine_name: the name of the engine to use.

    Returns a list of document types.
    """
    try:
        engine = app.data.engines[engine_name]

        if not isinstance(engine, MappingEngine):
            response = jsonify({
                'error': {
                    'desc': 'engine does not support mapping',
                    'key': engine_name,
                    'code': 'EngineNotSupported',
                }
            })
            response.status_code = 501
            return response

    except KeyError:
        response = jsonify({
            'error': {
                'desc': 'engine not found',
                'key': engine_name,
                'code': 'EngineNotFound',
            }
        })
        response.status_code = 404
        return response

    data = engine.get_doc_types()
    response = jsonify(data)
    response.status_code = 200
    return response


@app.route('/<engine_name>/mapping/<doc_type>', methods=['GET'])
def get_mapping(engine_name: str, doc_type: str) -> Response:
    """
    Get a list of available doc types.

    :param engine_name: the name of the engine to use.

    Returns a list of document types.
    """
    try:
        engine = app.data.engines[engine_name]

        if not isinstance(engine, MappingEngine):
            response = jsonify({
                'error': {
                    'desc': 'engine does not support mapping',
                    'key': engine_name,
                    'code': 'EngineNotSupported',
                }
            })
            response.status_code = 501
            return response

    except KeyError:
        response = jsonify({
            'error': {
                'desc': 'engine not found',
                'key': engine_name,
                'code': 'EngineNotFound',
            }
        })
        response.status_code = 404
        return response

    data = engine.get_mapping(doc_type).to_dict()
    response = jsonify(data)
    response.status_code = 200
    return response


@app.errorhandler(QueryParserError)
def parser_error(ex: QueryParserError) -> Response:
    """
    Handle parse errors.
    """
    response = jsonify({
        'error': {
            'desc': str(ex),
            'code': 'QueryParserError',
        }
    })
    response.status_code = 400
    return response


@app.after_request
def app_cors(response: Response) -> Response:
    """
    CORS request handler.

    Adds the CORS headers to all responses so that the API can be used directly from
    a web browser.
    """
    headers = {
        'Access-Control-Allow-Origin': request.headers.get('Origin', '*'),
        'Access-Control-Allow-Credentials': 'true',
        'Access-Control-Allow-Methods': 'POST, OPTIONS, GET, PUT, DELETE, HEAD, TRACE',
        'Access-Control-Allow-Headers': request.headers.get(
            'Access-Control-Request-Headers', 'Authorization'
        ),
    }

    for k, v in headers.items():
        response.headers[k] = v

    if app.debug:
        response.headers['Access-Control-Max-Age'] = '1'

    return response


if __name__ == '__main__':
    start_app(app)



"""
Relevance facets module.

This module provides the interface definitions for defining facets.
"""

import abc
import dateutil.parser
from dateutil.relativedelta import relativedelta


class FacetDefinitionError(TypeError):
    """
    Exception class when a facet is not defined properly.
    """
    def __init__(self, name: str, msg: str):
        """
        Initialize the exception.

        :param name: the name of the facet that caused the error.
        :param msg: the error message.
        """
        super().__init__('{0}: {1}'.format(name, msg))
        self.name = name


class FacetValueError(ValueError, RuntimeError):
    """
    Exception class when a supplied facet value is invalid.
    """
    def __init__(self, name: str, value: str):
        """
        Initialize the exception.

        :param name: the name of the facet that caused the error.
        :param value: the value of the facet that caused the error.
        """
        super().__init__('unknown value "{0}" for facet "{1}"'.format(value, name))
        self.name = name
        self.value = value


class FacetNotDefinedError(NameError, RuntimeError):
    """
    Exception class for undefined facets.
    """
    def __init__(self, name: str):
        """
        Initialize the exception.

        :param name: the name of the facet that caused the error.
        """
        super().__init__('the facet "{0}" is not defined'.format(name))
        self.name = name


class Facet(object, metaclass=abc.ABCMeta):
    """
    Facet definition class.

    This class is used to define facets within an engine.
    """
    def __init__(self, *, field: str=None, path: str=None):
        """
        Initialize the definition.

        :param field: the field for the facet. Mutually exclusive with path.
        :param path: the path or the facet. Mutually exclusive with field.
        :param facet_type: the type of facet to define.
        """
        if field is not None and path is not None:
            raise FacetDefinitionError(None, 'field and path are mutually exclusive')

        self.field = field
        self.path = path


class TermFacet(Facet):
    """
    Facet type for term facets.

    Aggregates on matching terms.
    """
    pass


class HistogramFacet(Facet):
    """
    Facet type for histogram facets.

    Facets that use ranges should inherit from this abstract class.
    """
    @abc.abstractmethod
    def get_range_for(self, value: object) -> tuple:
        """
        Get a start and end range given a specific value.

        :param value: the value to get the range for.
        :returns: a tuple containing the start and end values for the range.
        """
        pass


class DateFacet(HistogramFacet):
    """
    Facet type for date/time histogram facets.

    Aggregates on date intervals.
    """
    # Interval definitions
    YEAR = 'year'
    QUARTER = 'quarter'
    MONTH = 'month'
    WEEK = 'week'
    DAY = 'day'
    HOUR = 'hour'
    MINUTE = 'minute'
    SECOND = 'second'

    def __init__(self, *args, interval: str=MONTH, **kwargs):
        """
        Initialize the facet.

        :param interval: the interval at which to group the results: year, quarter, month,
        week, day, hour, minute, second.
        """
        super().__init__(*args, **kwargs)
        self.interval = interval

    def get_range_for(self, value: int) -> tuple:
        """
        Override abstract method.
        """
        start = dateutil.parser.parse(value)
        prop, delta = self.interval, 1

        if self.interval == DateFacet.YEAR:
            if start.month != 1 or start.day != 1 or start.hour != 0 or \
               start.minute != 0 or start.second != 0:
                raise FacetValueError(self, value)

        if self.interval == DateFacet.QUARTER:
            if start.month not in [1, 4, 7, 10] or start.day != 1 or \
               start.hour != 0 or start.minute != 0 or start.second != 0:
                raise FacetValueError(self, value)
            prop, delta = 'month', 3

        if self.interval == DateFacet.MONTH:
            if start.day != 1 or start.hour != 0 or \
               start.minute != 0 or start.second != 0:
                raise FacetValueError(self, value)

        if self.interval == DateFacet.WEEK:
            if start.weekday() != 0:
                raise FacetValueError(self, value)
            prop, delta = 'day', 7

        if self.interval == DateFacet.DAY:
            if start.hour != 0 or start.minute != 0 or start.second != 0:
                raise FacetValueError(self, value)

        if self.interval == DateFacet.HOUR:
            if start.minute != 0 or start.second != 0:
                raise FacetValueError(self, value)

        if self.interval == DateFacet.MINUTE:
            if start.second != 0:
                raise FacetValueError(self, value)

        end = start + relativedelta(**{'{0}s'.format(prop): delta})

        return start.isoformat(), end.isoformat()


class IntervalFacet(HistogramFacet):
    """
    Facet type for histogram facets.

    Aggregates on numeric intervals.
    """
    def __init__(self, *args, interval: int, **kwargs):
        """
        Initialize the facet.

        :param interval: the interval at which to group the results.
        """
        super().__init__(*args, **kwargs)
        self.interval = interval

    def get_range_for(self, value: int) -> tuple:
        """
        Override abstract method.
        """
        if (value % self.interval) > 0:
            raise FacetValueError(self, value)

        return value, value + self.interval


class RangeFacet(HistogramFacet):
    """
    Facet type for rangeed histogram facets.

    Aggregates on specific numeric intervals.
    """
    def __init__(self, *args, ranges: dict, **kwargs):
        """
        Initialize the facet.

        :param interval: the interval at which to group the results.
        """
        super().__init__(*args, **kwargs)
        self.ranges = ranges

    def get_range_for(self, value: int) -> tuple:
        """
        Override abstract method.
        """
        try:
            start, end = self.ranges[value]
        except KeyError:
            raise FacetValueError(self, value)

        return start, end



"""
Content pipeline manager.

Relevance is a software suite that aims to address the common and recurring problem
of ingesting content, delivering it, searching it with relevant results and analyzing
its usage by the end-users, acting on the analytic data to provide the best results
possible.
"""

# Package version
__version__ = '0.2.0'


class Document(dict):
    """
    Document class.

    This class represents a document that can be searched for or indexed using an engine
    implementation.
    """
    def __init__(self, uid: object, doc_type: str):
        """
        Initialize the document.

        :param uid: the unique document identifier.
        :param doc_type: the document type, usually a table name or similar.
        """
        self.uid = uid
        self.doc_type = doc_type

    def __eq__(self, other):
        """
        Overload operator for comparison.
        """
        return dict(self) == dict(other) and self.uid == other.uid and \
            self.doc_type == other.doc_type



"""
Relevance Elastic engine.

This module provides the storage and search engine connector for ElasticSearch instances.
"""

import requests
import datetime
from collections import OrderedDict

from relevance.query import Search
from relevance.query import LogicalOperator
from relevance.query import ComparisonOperator
from relevance.facet import TermFacet
from relevance.facet import HistogramFacet
from relevance.facet import DateFacet
from relevance.facet import IntervalFacet
from relevance.facet import RangeFacet
from relevance.facet import FacetNotDefinedError
from relevance.engine import SearchEngine
from relevance.engine import SearchRuntimeError
from relevance.engine import Result
from relevance.engine import ResultSet
from relevance.engine import MappingEngine
from relevance.engine import MappingEngineError
from relevance.mapping import Mapping
from relevance.mapping import Field
from relevance.mapping import ObjectField


class ElasticSearchEngine(SearchEngine, MappingEngine):
    """
    ElasticSearch storage engine.

    This class provides the storage and search engine for ElasticSearch.
    """
    def __init__(self, *args, **kwargs):
        """
        Initialize the engine.
        """
        super().__init__(*args, **kwargs)

    def to_dsl(self, search: Search) -> dict:
        """
        Transform a search object into a query DSL.

        :param search: the input search object.
        :returns: the resulting ElasticSearch query DSL.
        """
        def build_query(query):
            """
            Build a query into a DSL.
            """
            result = []

            # Logical operator
            if query.logic == LogicalOperator.AND:
                operator = 'must'
            elif query.logic == LogicalOperator.OR:
                operator = 'should'
            elif query.logic == LogicalOperator.NOT:
                operator = 'must_not'
            else:
                raise TypeError('unknown operator {0}'.format(query.logic))

            # Query terms
            for x in query.terms:
                result.append({'match': {'_all': x}})

            # Query filters
            for field, op, value in query.facets:
                start, end = value, value
                part = ref = {}

                # Manual facet definition
                try:
                    facet_def = self.facets[field]

                    if facet_def.path is not None:
                        field = facet_def.path

                        path = []
                        for x in facet_def.path.split('.')[:-1]:
                            path.append(x)
                            ref.update({
                                'nested': {
                                    'path': '.'.join(path),
                                    'query': {},
                                },
                            })
                            ref = ref['nested']['query']
                    elif facet_def.field is not None:
                        field = facet_def.field

                    if isinstance(facet_def, HistogramFacet):
                        start, end = facet_def.get_range_for(value)

                # Auto facet definition
                except (KeyError, TypeError):
                    if not self.auto_filters:
                        raise FacetNotDefinedError(field)

                # Filter operator
                if value is None:
                    if op == ComparisonOperator.EQ:
                        ref.update({'missing': {'field': field}})
                    else:
                        ref.update({'exists': {'field': field}})
                elif op == ComparisonOperator.EQ and start == end:
                    ref.update({'term': {field: value}})
                elif op == ComparisonOperator.NEQ and start == end:
                    ref.update({'bool': {'must_not': {'term': {field: value}}}})
                elif op == ComparisonOperator.EQ and start != end:
                    ref.update({'range': {field: {'gte': start, 'lt': end}}})
                elif op == ComparisonOperator.NEQ and start != end:
                    ref.update({'range': {field: {'lt': start, 'gte': end}}})
                elif op == ComparisonOperator.GT:
                    ref.update({'range': {field: {'gt': start}}})
                elif op == ComparisonOperator.LT:
                    ref.update({'range': {field: {'lt': end}}})
                elif op == ComparisonOperator.GTE:
                    ref.update({'range': {field: {'gte': start}}})
                elif op == ComparisonOperator.LTE:
                    ref.update({'range': {field: {'lte': end}}})
                else:
                    raise TypeError('unknown operator {0}'.format(op))

                result.append(part)

            # Sub queries
            for x in query.queries:
                result.append(build_query(x))

            dsl = {'bool': {operator: result}}
            if operator == 'should':
                dsl['bool']['minimum_should_match'] = 1

            return dsl

        def build_facets(search):
            """
            Build a search object into a aggregations DSL.
            """
            lst = list(self.facets) if search.facets is None else search.facets
            aggs = {}

            for field in lst:
                part = {field: {}}
                ref = part[field]

                # Manual facet definition
                try:
                    facet_def = self.facets[field]

                    if facet_def.path is not None:
                        field = facet_def.path

                        path = []
                        for x in facet_def.path.split('.')[:-1]:
                            path.append(x)
                            ref.update({
                                'nested': {
                                    'path': '.'.join(path),
                                },
                                'aggs': {'_nested': {}},
                            })
                            ref = ref['aggs']['_nested']
                    elif facet_def.field is not None:
                        field = facet_def.field

                # Auto facet definition
                except (KeyError, TypeError):
                    if not self.auto_filters:
                        raise FacetNotDefinedError(field)

                    facet_def = TermFacet(field)

                if isinstance(facet_def, TermFacet):
                    ref.update({
                        'terms': {'field': field}
                    })

                elif isinstance(facet_def, DateFacet):
                    ref.update({
                        'date_histogram': {
                            'field': field,
                            'interval': facet_def.interval,
                        }
                    })

                elif isinstance(facet_def, IntervalFacet):
                    ref.update({
                        'histogram': {
                            'field': field,
                            'interval': facet_def.interval,
                        }
                    })

                elif isinstance(facet_def, RangeFacet):
                    ref.update({
                        'range': {
                            'field': field,
                            'keyed': True,
                            'ranges': [{'key': k, 'from': v[0], 'to': v[1]}
                                       for k, v in facet_def.ranges.items()],
                        },
                    })

                else:
                    pass

                aggs.update(part)

            if len(aggs) > 0:
                return aggs

        dsl = {}

        if search.queries is not None:
            dsl.update({
                'query': build_query(search.queries),
            })

        if search.slices:
            dsl.update({
                'from': search.slices[0],
                'size': search.slices[1],
            })

        if search.sorts:
            dsl.update({
                'sort': dict([(k, v.value) for k, v in search.sorts]),
            })

        if search.facets is None or len(search.facets) > 0:
            aggs = build_facets(search)
            if aggs is not None:
                dsl.update({
                    'aggs': aggs,
                })

        return dsl

    def search(self, search: Search) -> ResultSet:
        """
        Perform a search request on the engine.

        This method automatically transforms the search object into a query DSL, sends it
        and parses the result as a resultset object.
        """
        super().search(search)

        url = '{0}/{1}_search'.format(
            self.target,
            ','.join(search.doc_types) + '/' if search.doc_types is not None else '',
        )
        payload = self.to_dsl(search)

        try:
            results = ResultSet(search, self)
            response = requests.post(url, json=payload).json()
            results.time_end = datetime.datetime.now()
            results.count = response['hits']['total']
        except Exception as ex:
            raise SearchRuntimeError(str(ex))

        for x in response['hits']['hits']:
            result = Result(x['_id'], x['_type'], score=x['_score'])
            result.update(x['_source'])
            results.append(result)

        if 'aggregations' in response:
            results.facets = {}
            for name, data in response['aggregations'].items():
                results.facets[name] = OrderedDict()

                ref = data
                while '_nested' in ref:
                    ref = ref['_nested']

                if isinstance(ref['buckets'], dict):
                    for x, y in ref['buckets'].items():
                        results.facets[name][x] = y['doc_count']
                elif isinstance(ref['buckets'], list):
                    for x in ref['buckets']:
                        key = x.get('key_as_string', x.get('key'))
                        results.facets[name][key] = x['doc_count']

        return results

    def get_doc_types(self) -> list:
        """
        Get a list of available document types.
        """
        url = '{0}/_mapping'.format(self.target)

        try:
            response = requests.get(url).json()
        except Exception as ex:
            raise MappingEngineError(str(ex))

        result = set()
        for index in response:
            try:
                for doc_type in response[index]['mappings']:
                    result.add(doc_type)
            except KeyError:
                pass

        return list(result)

    def get_mapping(self, doc_type: str) -> Mapping:
        """
        Get the mapping object for a specific document type.
        """
        url = '{0}/_mapping'.format(self.target)

        try:
            response = requests.get(url).json()
        except Exception as ex:
            raise MappingEngineError(str(ex))

        for index in response:
            def build_mapping(properties):
                """
                Build a mapping object from a properties dict.
                """
                mapping = Mapping()

                for key, conf in properties.items():
                    options = dict(
                        stored=conf.get('store', False),
                        indexed=conf.get('index', False),
                    )

                    if conf['type'] == 'string':
                        field = Field(key, str, **options)
                    if conf['type'] in ('long', 'integer'):
                        field = Field(key, int, **options)
                    if conf['type'] == 'double':
                        field = Field(key, float, **options)
                    if conf['type'] == 'boolean':
                        field = Field(key, bool, **options)
                    if conf['type'] == 'object':
                        field = Field(key, dict, **options)
                    if conf['type'] == 'binary':
                        field = Field(key, bytes, **options)
                    if conf['type'] == 'date':
                        field = Field(key, datetime.datetime, **options)
                    if conf['type'] == 'nested':
                        field = ObjectField(key, build_mapping(conf['properties']))

                    mapping = mapping.add(field)

                return mapping

            try:
                properties = response[index]['mappings'][doc_type]['properties']
            except KeyError:
                continue

            return build_mapping(properties)

        raise MappingEngineError('no mapping for {0}'.format(doc_type))

    def put_mapping(self, doc_type: str, mapping: Mapping):
        """
        Update the mapping for a specific document type.
        """
        pass



"""
Relevance engine core.

This package provides a unified interface for interacting with the different storage
and search backends.
"""

import abc
import datetime

from relevance import Document
from relevance.query import Search
from relevance.mapping import Mapping


class EngineError(Exception):
    """
    Exception class for engine error.

    This is the base exception class when an engine implementation encounters an
    error during a request.
    """
    pass


class MappingEngineError(EngineError):
    """
    Exception class for mapping engines.
    """
    pass


class SearchEngineError(EngineError):
    """
    Exception class for search engines.
    """
    pass


class SearchRuntimeError(SearchEngineError, RuntimeError):
    """
    Exception class when the search triggers an error at runtime.
    """
    pass


class MappingEngine(object, metaclass=abc.ABCMeta):
    """
    Abstract class for engines that support mapping.

    This class provides the interface for storage and search engines to support
    mapping definitions.
    """
    @abc.abstractmethod
    def get_doc_types(self) -> list:
        """
        Get a list of available document types.

        :returns: a list of document type strings.
        """
        pass

    @abc.abstractmethod
    def get_mapping(self, doc_type: str) -> Mapping:
        """
        Get the mapping object for a specific document type.

        :param doc_type: the document type.
        :returns: the mapping object.
        """
        pass

    @abc.abstractmethod
    def put_mapping(self, doc_type: str, mapping: Mapping):
        """
        Update the mapping for a specific document type.

        :param doc_type: the document type.
        :param mapping: the mapping object.
        """
        pass


class SearchEngine(object, metaclass=abc.ABCMeta):
    """
    Search Engine abstract class.

    This class provides the abstract interface for storage engines to implement search
    functionality.
    """
    def __init__(self, target: str, *, facets: dict=None, auto_filters: bool=True):
        """
        Initialize the engine.

        :param target: the storage engine target.
        :param facets: a dictionary of facets definitions.
        :param auto_filters: whether to enable automatic filter definitions.
        """
        self.target = target
        self.facets = facets if facets is not None else {}
        self.auto_filters = auto_filters

    @abc.abstractmethod
    def search(self, search: Search) -> 'ResultSet':
        """
        Perform a search request on the engine.

        :param request: the request to execute.
        :returns: a result set object.
        """
        pass


class Result(Document):
    """
    Result class.

    This class provides the interface for handling single search results.
    """
    def __init__(self, *args, score: float):
        """
        Initialize the result.

        :param score: the relevancy score.
        """
        super().__init__(*args)
        self.score = score


class ResultSet(list):
    """
    Result set class.

    This class provides the interface for handling search result lists.
    """
    def __init__(self, search: Search, engine: SearchEngine):
        """
        Initialize the result set.

        :param search: the search object.
        :param engine: the engine object.
        """
        self.search = search
        self.engine = engine
        self.time_start = datetime.datetime.now()
        self.time_end = None
        self.facets = None
        self.count = 0

    @property
    def time(self) -> float:
        """
        Get the total amount of time the request took.
        """
        return (self.time_end - self.time_start).total_seconds()



"""
Relevance API package.

This package provides some common interfaces for the different REST APIs to use.
"""

from flask import Flask


def start_app(app: Flask):
    """
    Start an API application.

    app -- the application object.
    """
    app.run(debug=True)
